@@ -2,14 +2,18 @@ from collections.abc import Iterable
 from typing import Any
 
 from sapiopylib.rest.DataRecordManagerService import DataRecordManager
+from sapiopylib.rest.User import SapioUser
 from sapiopylib.rest.pojo.DataRecord import DataRecord
 from sapiopylib.rest.pojo.DataRecordPaging import DataRecordPojoPageCriteria
 from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
+from sapiopylib.rest.utils.autopaging import QueryDataRecordsAutoPager, QueryDataRecordByIdListAutoPager, \
+    QueryAllRecordsOfTypeAutoPager
 from sapiopylib.rest.utils.recordmodel.PyRecordModel import PyRecordModel
 from sapiopylib.rest.utils.recordmodel.RecordModelManager import RecordModelManager, RecordModelInstanceManager, \
     RecordModelRelationshipManager
-from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType
-from sapiopylib.rest.utils.recordmodel.RelationshipPath import RelationshipPath, RelationshipPathDir
+from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType, WrappedRecordModel
+from sapiopylib.rest.utils.recordmodel.RelationshipPath import RelationshipPath, RelationshipNode, \
+    RelationshipNodeType
 
 from sapiopycommons.general.aliases import RecordModel, SapioRecord, FieldMap
 from sapiopycommons.general.custom_report_util import CustomReportUtil
@@ -21,19 +25,19 @@ class RecordHandler:
     """
     A collection of shorthand methods for dealing with the various record managers.
     """
-    __context: SapioWebhookContext
+    user: SapioUser
     dr_man: DataRecordManager
     rec_man: RecordModelManager
     inst_man: RecordModelInstanceManager
     rel_man: RecordModelRelationshipManager
 
-    def __init__(self, context: SapioWebhookContext):
+    def __init__(self, context: SapioWebhookContext | SapioUser):
         """
-        :param context: The current webhook context.
+        :param context: The current webhook context or a user object to send requests from.
         """
-        self.__context = context
-        self.dr_man = context.data_record_manager
-        self.rec_man = RecordModelManager(context.user)
+        self.user = context if isinstance(context, SapioUser) else context.user
+        self.dr_man = DataRecordManager(self.user)
+        self.rec_man = RecordModelManager(self.user)
         self.inst_man = self.rec_man.instance_manager
         self.rel_man = self.rec_man.relationship_manager
 
@@ -88,8 +92,9 @@ class RecordHandler:
         :return: The record models for the queried records and the final paging criteria.
         """
         dt: str = wrapper_type.get_wrapper_data_type_name()
-        records, paging_criteria = self.__exhaust_query_pages(dt, field, list(value_list), paging_criteria, page_limit)
-        return self.wrap_models(records, wrapper_type), paging_criteria
+        pager = QueryDataRecordsAutoPager(dt, field, list(value_list), self.user, paging_criteria)
+        pager.max_page = page_limit
+        return self.wrap_models(pager.get_all_at_once(), wrapper_type), pager.next_page_criteria
 
     def query_models_by_id(self, wrapper_type: type[WrappedType], ids: Iterable[int],
                            page_limit: int | None = None) -> list[WrappedType]:
@@ -120,8 +125,9 @@ class RecordHandler:
         :return: The record models for the queried records and the final paging criteria.
         """
         dt: str = wrapper_type.get_wrapper_data_type_name()
-        records, paging_criteria = self.__exhaust_query_id_pages(dt, list(ids), paging_criteria, page_limit)
-        return self.wrap_models(records, wrapper_type), paging_criteria
+        pager = QueryDataRecordByIdListAutoPager(dt, list(ids), self.user, paging_criteria)
+        pager.max_page = page_limit
+        return self.wrap_models(pager.get_all_at_once(), wrapper_type), pager.next_page_criteria
 
     def query_all_models(self, wrapper_type: type[WrappedType], page_limit: int | None = None) -> list[WrappedType]:
         """
@@ -149,8 +155,9 @@ class RecordHandler:
         :return: The record models for the queried records and the final paging criteria.
         """
         dt: str = wrapper_type.get_wrapper_data_type_name()
-        records, paging_criteria = self.__exhaust_query_all_pages(dt, paging_criteria, page_limit)
-        return self.wrap_models(records, wrapper_type), paging_criteria
+        pager = QueryAllRecordsOfTypeAutoPager(dt, self.user, paging_criteria)
+        pager.max_page = page_limit
+        return self.wrap_models(pager.get_all_at_once(), wrapper_type), pager.next_page_criteria
 
     def query_models_by_report(self, wrapper_type: type[WrappedType],
                                report_name: str,
@@ -169,7 +176,7 @@ class RecordHandler:
         :param page_limit: The maximum number of pages to query. If None, exhausts all possible pages.
         :return: The record models for the queried records.
         """
-        results: list[dict[str, Any]] = CustomReportUtil.run_system_report(self.__context, report_name, filters, page_limit)
+        results: list[dict[str, Any]] = CustomReportUtil.run_system_report(self.user, report_name, filters, page_limit)
         # Using the bracket operators because we want to throw an exception if RecordId doesn't exist in the report.
         ids: list[int] = [row["RecordId"] for row in results]
         return self.query_models_by_id(wrapper_type, ids)
@@ -391,6 +398,88 @@ class RecordHandler:
                 by_children.setdefault(child, []).append(record)
         return by_children
 
+    @staticmethod
+    def map_to_forward_side_link(models: Iterable[WrappedRecordModel], field_name: str,
+                                 side_link_type: type[WrappedType]) -> dict[WrappedRecordModel, WrappedType]:
+        """
+        Map a list of record models to their forward side link. The forward side link must already be loaded.
+
+        :param models: A list of record models.
+        :param field_name: The field name on the record models where the side link is located.
+        :param side_link_type: The record model wrapper of the forward side link.
+        :return: A dict[ModelType, SlideLink]. If an input model doesn't have a forward side link of the given type,
+            then it will map to None.
+        """
+        return_dict: dict[WrappedRecordModel, WrappedType] = {}
+        for model in models:
+            return_dict[model] = model.get_forward_side_link(field_name, side_link_type)
+        return return_dict
+
+    @staticmethod
+    def map_by_forward_side_link(models: Iterable[WrappedRecordModel], field_name: str,
+                                 side_link_type: type[WrappedType]) -> dict[WrappedType, list[WrappedRecordModel]]:
+        """
+        Take a list of record models and map them by their forward side link. Essentially an inversion of
+        map_to_forward_side_link. Input models that share a forward side link will end up in the same list.
+        The forward side link must already be loaded.
+
+        :param models: A list of record models.
+        :param field_name: The field name on the record models where the side link is located.
+        :param side_link_type: The record model wrapper of the forward side links.
+        :return: A dict[SideLink, list[ModelType]]. If an input model doesn't have a forward side link of the given type
+            pointing to it, then it will not be in the resulting dictionary.
+        """
+        to_side_link: dict[RecordModel, WrappedType] = RecordHandler\
+            .map_to_forward_side_link(models, field_name, side_link_type)
+        by_side_link: dict[WrappedType, list[RecordModel]] = {}
+        for record, side_link in to_side_link.items():
+            if side_link is None:
+                continue
+            by_side_link.setdefault(side_link, []).append(record)
+        return by_side_link
+
+    @staticmethod
+    def map_to_reverse_side_links(models: Iterable[WrappedRecordModel], field_name: str,
+                                  side_link_type: type[WrappedType]) -> dict[RecordModel, list[WrappedRecordModel]]:
+        """
+        Map a list of record models to a list reverse side links of a given type. The reverse side links must already
+        be loaded.
+
+        :param models: A list of record models.
+        :param field_name: The field name on the side linked model where the side link to the given record models is
+            located.
+        :param side_link_type: The record model wrapper of the reverse side links.
+        :return: A dict[ModelType, list[SideLink]]. If an input model doesn't have reverse side links of the given type,
+            then it will map to an empty list.
+        """
+        return_dict: dict[WrappedRecordModel, list[WrappedType]] = {}
+        for model in models:
+            return_dict[model] = model.get_reverse_side_link(field_name, side_link_type)
+        return return_dict
+
+    @staticmethod
+    def map_by_reverse_side_links(models: Iterable[WrappedRecordModel], field_name: str,
+                                  side_link_type: type[WrappedType]) -> dict[WrappedType, list[WrappedRecordModel]]:
+        """
+        Take a list of record models and map them by their reverse side links. Essentially an inversion of
+        map_to_reverse_side_links. Input models that share a reverse side link will end up in the same list.
+        The reverse side links must already be loaded.
+
+        :param models: A list of record models.
+        :param field_name: The field name on the side linked model where the side link to the given record models is
+            located.
+        :param side_link_type: The record model wrapper of the reverse side links.
+        :return: A dict[SideLink, list[ModelType]]. If an input model doesn't have reverse side links of the given type
+            pointing to it, then it will not be in the resulting dictionary.
+        """
+        to_side_links: dict[RecordModel, list[WrappedType]] = RecordHandler\
+            .map_to_reverse_side_links(models, field_name, side_link_type)
+        by_side_links: dict[WrappedType, list[RecordModel]] = {}
+        for record, side_links in to_side_links.items():
+            for side_link in side_links:
+                by_side_links.setdefault(side_link, []).append(record)
+        return by_side_links
+
     @staticmethod
     def map_by_id(models: Iterable[SapioRecord]) -> dict[int, SapioRecord]:
         """
@@ -518,7 +607,7 @@ class RecordHandler:
         # Otherwise, create a new fields map list.
         return [{field_name: value} for value in values]
 
-    # FR-46155: Update relationship path traversing functions to be non-const and take in a wrapper type so that the
+    # FR-46155: Update relationship path traversing functions to be non-static and take in a wrapper type so that the
     # output can be wrapped instead of requiring the user to wrap the output.
     def get_linear_path(self, models: Iterable[RecordModel], path: RelationshipPath, wrapper_type: type[WrappedType]) \
             -> dict[RecordModel, WrappedType | None]:
@@ -527,6 +616,8 @@ class RecordHandler:
         path, if any. The hierarchy must be linear (1:1 relationship between data types at every step) and the
         relationship path must already be loaded.
 
+        Currently, the relationship path may only contain parent/child nodes.
+
         :param models: A list of record models.
         :param path: The relationship path to follow.
         :param wrapper_type: The record model wrapper to use.
@@ -534,16 +625,20 @@ class RecordHandler:
             path couldn't be reached, the record will map to None.
         """
         ret_dict: dict[RecordModel, WrappedType | None] = {}
-        path: list[tuple[RelationshipPathDir, str]] = path.path
+        # PR-46832: Update path traversal to account for changes to RelationshipPath in Sapiopylib.
+        path: list[RelationshipNode] = path.path
         for model in models:
             current: PyRecordModel = model if isinstance(model, PyRecordModel) else model.backing_model
-            for direction, datatype in path:
+            for node in path:
+                direction = node.direction
                 if current is None:
                     break
-                if direction == RelationshipPathDir.CHILD:
-                    current = current.get_child_of_type(datatype)
-                elif direction == RelationshipPathDir.PARENT:
-                    current = current.get_parent_of_type(datatype)
+                if direction == RelationshipNodeType.CHILD:
+                    current = current.get_child_of_type(node.data_type_name)
+                elif direction == RelationshipNodeType.PARENT:
+                    current = current.get_parent_of_type(node.data_type_name)
+                else:
+                    raise SapioException("Unsupported path direction.")
             ret_dict.update({model: self.inst_man.wrap(current, wrapper_type) if current else None})
         return ret_dict
 
@@ -554,6 +649,8 @@ class RecordHandler:
         path, if any. The hierarchy may be non-linear (1:Many relationships between data types are allowed) and the
         relationship path must already be loaded.
 
+        Currently, the relationship path may only contain parent/child nodes.
+
         :param models: A list of record models.
         :param path: The relationship path to follow.
         :param wrapper_type: The record model wrapper to use.
@@ -561,19 +658,23 @@ class RecordHandler:
             path couldn't be reached, the record will map to an empty list.
         """
         ret_dict: dict[RecordModel, list[WrappedType]] = {}
-        path: list[tuple[RelationshipPathDir, str]] = path.path
+        # PR-46832: Update path traversal to account for changes to RelationshipPath in Sapiopylib.
+        path: list[RelationshipNode] = path.path
         for model in models:
             current_search: set[PyRecordModel] = {model if isinstance(model, PyRecordModel) else model.backing_model}
             next_search: set[PyRecordModel] = set()
             # Exhaust the records at each step in the path, then use those records for the next step.
-            for direction, datatype in path:
+            for node in path:
+                direction = node.direction
                 if len(current_search) == 0:
                     break
                 for search in current_search:
-                    if direction == RelationshipPathDir.CHILD:
-                        next_search.update(search.get_children_of_type(datatype))
-                    elif direction == RelationshipPathDir.PARENT:
-                        next_search.update(search.get_parents_of_type(datatype))
+                    if direction == RelationshipNodeType.CHILD:
+                        next_search.update(search.get_children_of_type(node.data_type_name))
+                    elif direction == RelationshipNodeType.PARENT:
+                        next_search.update(search.get_parents_of_type(node.data_type_name))
+                    else:
+                        raise SapioException("Unsupported path direction.")
                 current_search = next_search
                 next_search = set()
             ret_dict.update({model: self.inst_man.wrap_list(list(current_search), wrapper_type)})
@@ -592,6 +693,8 @@ class RecordHandler:
         relationships (e.g. a sample with is aliquoted to a number of samples, then those aliquots are pooled back
         together into a single sample).
 
+        Currently, the relationship path may only contain parent/child nodes.
+
         :param models: A list of record models.
         :param path: The relationship path to follow.
         :param wrapper_type: The record model wrapper to use.
@@ -599,16 +702,20 @@ class RecordHandler:
             path couldn't be reached, the record will map to None.
         """
         ret_dict: dict[RecordModel, WrappedType | None] = {}
-        path: list[tuple[RelationshipPathDir, str]] = path.path
+        # PR-46832: Update path traversal to account for changes to RelationshipPath in Sapiopylib.
+        path: list[RelationshipNode] = path.path
         for model in models:
             current: list[PyRecordModel] = [model if isinstance(model, PyRecordModel) else model.backing_model]
-            for direction, datatype in path:
+            for node in path:
+                direction = node.direction
                 if len(current) == 0:
                     break
-                if direction == RelationshipPathDir.CHILD:
-                    current = current[0].get_children_of_type(datatype)
-                elif direction == RelationshipPathDir.PARENT:
-                    current = current[0].get_parents_of_type(datatype)
+                if direction == RelationshipNodeType.CHILD:
+                    current = current[0].get_children_of_type(node.data_type_name)
+                elif direction == RelationshipNodeType.PARENT:
+                    current = current[0].get_parents_of_type(node.data_type_name)
+                else:
+                    raise SapioException("Unsupported path direction.")
             ret_dict.update({model: self.inst_man.wrap(current[0], wrapper_type) if current else None})
         return ret_dict
 
@@ -616,43 +723,6 @@ class RecordHandler:
                               paging_criteria: DataRecordPojoPageCriteria | None,
                               page_limit: int | None) \
             -> tuple[list[DataRecord], DataRecordPojoPageCriteria | None]:
-        has_next_page: bool = True
-        records: list[DataRecord] = []
-        cur_page: int = 1
-        while has_next_page and (not page_limit or cur_page < page_limit):
-            page_result = self.dr_man.query_data_records(data_type_name, field, value_list, paging_criteria)
-            paging_criteria = page_result.next_page_criteria
-            has_next_page = page_result.is_next_page_available
-            records.extend(page_result.result_list)
-            cur_page += 1
-        return records, paging_criteria
-
-    def __exhaust_query_id_pages(self, data_type_name: str, id_list: list[int],
-                                 paging_criteria: DataRecordPojoPageCriteria | None,
-                                 page_limit: int | None) \
-            -> tuple[list[DataRecord], DataRecordPojoPageCriteria | None]:
-        has_next_page: bool = True
-        records: list[DataRecord] = []
-        cur_page: int = 1
-        while has_next_page and (not page_limit or cur_page < page_limit):
-            page_result = self.dr_man.query_data_records_by_id(data_type_name, id_list, paging_criteria)
-            paging_criteria = page_result.next_page_criteria
-            has_next_page = page_result.is_next_page_available
-            records.extend(page_result.result_list)
-            cur_page += 1
-        return records, paging_criteria
-
-    def __exhaust_query_all_pages(self, data_type_name: str,
-                                  paging_criteria: DataRecordPojoPageCriteria | None,
-                                  page_limit: int | None) \
-            -> tuple[list[DataRecord], DataRecordPojoPageCriteria | None]:
-        has_next_page: bool = True
-        records: list[DataRecord] = []
-        cur_page: int = 1
-        while has_next_page and (not page_limit or cur_page < page_limit):
-            page_result = self.dr_man.query_all_records_of_type(data_type_name, paging_criteria)
-            paging_criteria = page_result.next_page_criteria
-            has_next_page = page_result.is_next_page_available
-            records.extend(page_result.result_list)
-            cur_page += 1
-        return records, paging_criteria
+        pager = QueryDataRecordsAutoPager(data_type_name, field, value_list, self.user, paging_criteria)
+        pager.max_page = page_limit
+        return pager.get_all_at_once(), pager.next_page_criteria
